/* slk.c: Palm Serial Link Protocol routines for tcpusb.

   Copyright 2005 Ton van Overbeek.

   This is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.  */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <usb.h>

/* External functions */

int write_usb (char *buff, int size, int timeout);
int read_usb (char *buff, int size, int timeout);
void logtraffic (FILE *logfile, char direction, char *buff, int soze);

/* External global variables */

extern int verbose;

/* Portable memory access macros */

#define get_long(ptr) ((((unsigned char*)(ptr))[0] << 24) | \
                       (((unsigned char*)(ptr))[1] << 16) | \
                       (((unsigned char*)(ptr))[2] << 8)  | \
                       (((unsigned char*)(ptr))[3] << 0))

#define get_treble(ptr) ((((unsigned char*)(ptr))[0] << 16) | \
                         (((unsigned char*)(ptr))[1] << 8)  | \
                         (((unsigned char*)(ptr))[2] << 0))

#define get_short(ptr) ((((unsigned char*)(ptr))[0] << 8)  | \
                        (((unsigned char*)(ptr))[1] << 0))

#define get_byte(ptr) (((unsigned char*)(ptr))[0])

#define set_long(ptr,val) ((((unsigned char*)(ptr))[0] = ((val) >> 24) & 0xff), \
                          (((unsigned char*)(ptr))[1] = ((val) >> 16) & 0xff), \
                          (((unsigned char*)(ptr))[2] = ((val) >> 8) & 0xff), \
                          (((unsigned char*)(ptr))[3] = ((val) >> 0) & 0xff))

#define set_treble(ptr,val) ((((unsigned char*)(ptr))[0] = ((val) >> 16) & 0xff), \
                             (((unsigned char*)(ptr))[1] = ((val) >> 8) & 0xff), \
                             (((unsigned char*)(ptr))[2] = ((val) >> 0) & 0xff))

#define set_short(ptr,val) ((((unsigned char*)(ptr))[0] = ((val) >> 8) & 0xff), \
                            (((unsigned char*)(ptr))[1] = ((val) >> 0) & 0xff))

#define set_byte(ptr,val) (((unsigned char*)(ptr))[0]=(val))

/* Following CRC code borrowed from xmodem.c */

#define CRC16 0x1021        /* Generator polynomial (X^16 + X^12 + X^5 + 1) */

static unsigned short *crctab;

/* Call this to init the fast CRC-16 calculation table.  */

static void
crcinit (void)
{
  static int crctab_inited = 0;
  int val;

  if (crctab_inited == 1)
    return;

  crctab = malloc (256 * sizeof (short));

  for (val = 0; val <= 255; val++)
    {
      int i;
      unsigned int crc;

      crc = val << 8;

      for (i = 0; i < 8; ++i)
    {
      crc <<= 1;

      if (crc & 0x10000)
        crc ^= CRC16;
    }

      crctab [val] = crc;
    }

  crctab_inited = 1;
}

/* Calculate a CRC-16 for the LEN byte message pointed at by P.  */

static unsigned short
docrc (unsigned char *p, int len)
{
  unsigned short crc = 0;

  crcinit ();
  while (len-- > 0)
    crc = (crc << 8) ^ crctab [(crc >> 8) ^ *p++];

  return crc;
}

/* Routines to perform RPC */

#define        sysPktMaxMemChunk    256
#define        sysPktMaxBodySize    (sysPktMaxMemChunk+16)

#define        sysErrorClass        0x0500
#define        sysErrParamErr       (sysErrorClass | 2)
#define        sysErrNoFreeRAM      (sysErrorClass | 4)

static void
rpc_inc_param_count (char* buffer)
{
  set_short (&buffer[12], get_short(&buffer[12]) + 1);
}

static int
rpc_prepare (char* buffer, int trapword)
{
  trapword |= 0xA000; /* Make double-dog sure this is set -- PalmSim requires it. */

  set_byte  (&buffer[0], 0x0A);          /* RPC command */
  set_byte  (&buffer[1], 0x00);          /* filler */
  set_short (&buffer[2], trapword);      /* trapword */
  set_long  (&buffer[4], 0);             /* D0 result */
  set_long  (&buffer[8], 0);             /* A0 result */
  set_short (&buffer[12], 0);            /* number of parameters */

  return 14;
}

static int
rpc_add_short_val (char* buffer, int offset, short val)
{
  set_byte  (&buffer[offset + 0], 0);    /* by ref == false */
  set_byte  (&buffer[offset + 1], 2);    /* parameter size */
  set_short (&buffer[offset + 2], val);  /* parameter data */
  rpc_inc_param_count (buffer);
  return offset + 4;
}

static int
rpc_add_long_val (char* buffer, int offset, long val)
{
  set_byte (&buffer[offset + 0], 0);    /* by ref == false */
  set_byte (&buffer[offset + 1], 4);    /* parameter size */
  set_long (&buffer[offset + 2], val);  /* parameter data */
  rpc_inc_param_count (buffer);
  return offset + 6;
}

static int
rpc_add_long_ref (char* buffer, int offset, long val)
{
  set_byte (&buffer[offset + 0], 1);    /* by ref == true */
  set_byte (&buffer[offset + 1], 4);    /* parameter size */
  set_long (&buffer[offset + 2], val);  /* parameter data */
  rpc_inc_param_count (buffer);
  return offset + 6;
}

static unsigned char transid = 0x11;

/* Send a packet to the remote machine.
   The data of the packet is in BUF.  */

static int
putpkt_to (char *buf, int len, int socket)
{
  static unsigned char buffer[0xffff];
  int i;

  buffer[0] = 0xBE;
  buffer[1] = 0xEF;
  buffer[2] = 0xED;
  buffer[3] = socket;
  buffer[4] = socket;
  buffer[5] = 0; /* slkPktTypeSystem */
  buffer[6] = len >> 8;
  buffer[7] = len;
  buffer[8] = ++transid;
  buffer[9] = 0;
  for (i=0;i<9;i++)
    buffer[9] += buffer[i];

  memcpy (buffer+10, buf, len);

  set_short (buffer+len+10, docrc(buffer, len+10));

  if (write_usb (buffer, len+12, 1000) != len+12) {
    fprintf (stderr, "putpkt_to: usb_bulk_write failed\n");
    return 1;
  }

  return 0;
}

static int
readchar (int timeout)
{
  static unsigned char buffer[128];
  static int lastinx = 0;
  static int currinx = 0;
  int ret;

  if (currinx >= lastinx) {
    if (verbose)
      printf ("readchar: currinx = %d, lastinx = %d, ", currinx, lastinx);
    ret = read_usb (buffer, 128, timeout ? 500 : 0);
    if (ret <= 0) {
      if (verbose)
        printf ("ret = %d\n", ret);
      return -2;
    } else {
      lastinx = ret;
      currinx = 0;
      if (verbose) {
        printf ("read %d bytes\n", ret);
        logtraffic (stdout, '<', buffer, ret);
      }
    }
  }

  return buffer[currinx++];
}

/* Read a packet from the remote machine, with error checking,
   and store it in BUF.  BUF is expected to be of size PBUFSIZ.
   If FOREVER, wait forever rather than timing out; this is used
   while the target is executing user code.  */

static int
getpkt (char **buf, int forever)
{
  static unsigned char buffer[0xffff];
  int state = 0;
  int c;
  int i;
  unsigned int len, csum, crc;

  while (1) {
    c = readchar (-1);
    if (c < 0) {
       if (!forever)
          return 0;
    }

    buffer[state] = c;

    switch (state) {
        case 0:
            if (c == 0xBE) state++; else state=0;
            break;
        case 1:
            if  (c == 0xEF) state++; else state=0;
            break;
        case 2:
            if  (c == 0xED) state++; else state=0;
            break;
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
            state++;
            break;
        case 9:
            csum = 0;
            for (i=0;i<9;i++)
              csum += buffer[i];
            if ((csum & 0xff) == c) {
              len = (buffer[6] << 8) | buffer[7];
              state++;
            }
            else
              state=0;
            break;
        default:
            if (state >= 10) {
              if (state < 10+len) {
                state++;
              }
              else if (state == 10+len) {
                crc = c;
                state++;
              } else if (state == 11+len) {
                unsigned long mycrc = docrc (buffer,len+10);
                crc = (crc<<8)|c;
                if ((crc & 0xffff) == mycrc) {
                   *buf = buffer;
                   return len+10;
                } else
                  state = 0;
              } else
                state = 0;
            } else
              state = 0;
       }

  }
}

int
rd_initial (void)
{
  char *buff;
  while (getpkt (&buff, 0) > 0) ;
  return 0;
}

void
send_FtrGet (int socket)
{
  char buffer[sysPktMaxBodySize];
  int offset;

  if (verbose) {
    printf("send_FtrGet: sending FtrGet RPC on %s socket\n",
           (socket == 0 ? "debug" : "console"));
  }
  offset = rpc_prepare (buffer, 0xA27B /* FtrGet */ );
  offset = rpc_add_long_ref (buffer, offset, 0x0L);
  offset = rpc_add_short_val (buffer,offset, 1 /* sysFtrNumRomVersion */ );
  offset = rpc_add_long_val (buffer, offset, 'psys' /* Creator */ );

  putpkt_to (buffer, offset, socket);
}

static char romstr[25];

int
rcv_FtrGet (void)
{
  char *buff;
  int ret;
  unsigned long Romversion;
  int maj, min, fix, stage, build;

  for (;;) {
    if ((ret = getpkt (&buff, 0)) == 0)
      return 0;
    if (buff[10] == (char)0x8A &&
        buff[12] == (char)0xA2 &&
        buff[13] == (char)0x7B &&
        get_long (&buff[14]) == 0x0L) {
      Romversion = get_long (&buff[26]);
      maj = (Romversion >> 24) & 0xff;
      min = (Romversion >> 20) & 0x0f;
      fix = (Romversion >> 16) & 0x0f;
      stage = (Romversion >> 12) & 0x0f;
      build = Romversion & 0x0fff;
      sprintf (romstr, "%d.%d", maj, min);
      if (fix) sprintf (&romstr[strlen(romstr)], ".%d", fix);
      if (stage < 3) {
        if (stage == 0) sprintf (&romstr[strlen(romstr)], "dev");
        else if (stage == 1) sprintf (&romstr[strlen(romstr)], "alpha");
        else if (stage == 2) sprintf (&romstr[strlen(romstr)], "beta");
        sprintf (&romstr[strlen(romstr)], "%d", build);
      }
      return ret;
    }
  }
}

void
send_SysReset (int socket)
{
  char buffer[sysPktMaxBodySize];
  int offset;

  if (verbose) {
     printf("send_SysReset: sending SysReset RPC on %s socket\n",
            (socket == 0 ? "debug" : "console"));
  }
  offset = rpc_prepare (buffer, 0xA08C /* SysReset */ );
  putpkt_to (buffer, offset, socket);
}

void
send_status_query (void)
{
  char buffer[2] = { 0, 0 };

  putpkt_to (buffer, 2, 0);
}

char *
getromstr (void)
{
  if (romstr[0])
    return romstr;
  return (char *)0;
}
