/* tcpusb.c: TCP-USB bridge for m68k-palmos-gdb

   Copyright 2005 Ton van Overbeek.

   This is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <popt.h>

/* External functions */

int setup_socket (int port);
int poll_socket (int sock);
void print_client_info (void);
int poll_usb (int *slkport);
void send_status_query ();
void init_usb (int debuglevel);
void close_usb (void);
int write_usb (char *buff, int size, int timeout);
int read_usb (char *buff, int size, int timeout);
void print_dev_info (void);
void logtraffic (FILE *logfile, char direction, char *buff, int size);
void send_SysReset (int slksocket);

/* Global variables */

int verbose = 0;

/* Handler for interrupt signal (Ctrl-C) */

void
ctrlc_handler(int signo)
{
  printf ("Interrupted ...\n"
          "Trying to send SysReset to device.\n");
  send_SysReset (0);
  send_SysReset (1);
  close_usb ();
  exit (EXIT_FAILURE);
}

/* The main  program */

int
main (int argc, const char *argv[])
{
  int debuglevel = 0;
  char *logfname = NULL;
  poptContext optCon;
  int rc;
  int tcpport = 2000;
  int srvsock, gdbsock;
  int slkport;
  int nr;
  int usbok = 0, tcpok = 0;
  char buff[256];
  FILE *logfile = (FILE *)0;

  struct poptOption optionsTable[] = {
    { "debug",   'd', POPT_ARG_INT,    &debuglevel, 0, "set debuglevel for libusb", "level" },
    { "log",     'l', POPT_ARG_STRING, &logfname,   0, "log all bridge traffic to file (use \'-\' for stdout)", "filename" },
    { "verbose", 'v', 0,               &verbose,    0, "be verbose during device handshake" },
    { "port",    'p', POPT_ARG_INT,    &tcpport,    0, "TCP port to listen on" },
    POPT_AUTOHELP
    POPT_TABLEEND
  };

  /* Greeting and version */
  printf("Tcpusb - A bridge between Palm OS USB devices and m68k-palmos-gdb\n"
         "Version: 0.2 2005-05-29\n");

  /* Option processing */
  optCon = poptGetContext (NULL, argc, argv, optionsTable, 0);
  rc = poptGetNextOpt (optCon);
  if (rc < -1) {
    /* an error occurred during option processing */
    fprintf (stderr, "%s: %s\n\n",
             poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
             poptStrerror(rc));
    poptPrintHelp (optCon, stderr, 0);
    return EXIT_FAILURE;
  }

  /* Print out options */
  printf ("Options chosen:\n");
  printf ("  debuglevel = %d\n", debuglevel);
  printf ("  log = %s\n", logfname ? logfname : "<None>");
  printf ("  verbose = %d\n", verbose);
  printf ("  port = %d\n\n", tcpport);

  poptFreeContext (optCon);

  /* Open log file */

  if (logfname) {
    if (!strcmp (logfname, "-")) {
      logfile = stdout;
    }
    else {
      logfile = fopen (logfname, "w");
      if (!logfile) {
        fprintf (stderr, "Cannot open logfile %s for writing\n", logfname);
        return EXIT_FAILURE;
      }
    }
  }

  /* Initialize USB side */
  init_usb (debuglevel);

  /* Initialize TCP side */
  srvsock = setup_socket (tcpport);

  /* Wait till both sides are connected */
  printf ("Waiting ...\n"); fflush (stdout);
  for (; !tcpok || !usbok ;) {

    if (usbok) {
      /* Check Palm USB device still there */
      nr = read_usb (buff, 64, 10);
      if (nr > 0 )
        printf ("Warning: Bridge not active yet, USB data ignored\n");
      else if (nr < 0 && nr != -ETIMEDOUT) {
        printf ("USB connection broken\n");
        close_usb ();
        exit (EXIT_FAILURE);
      }
    }
    else if (poll_usb (&slkport)) {
      signal (SIGINT, ctrlc_handler);
      printf ("PalmOS PDA connected via USB on %s port\n",
              slkport == 1 ? "console" : "debug");
      print_dev_info ();
      usbok = -1;
    }

    if (tcpok) {
      /* Check TCP connection still OK */
      nr = read (gdbsock, buff, 256);
      if (nr > 0)
        printf ("Warning: Bridge not active yet, tcp data ignored\n");
      else if (nr == 0 || errno != EWOULDBLOCK) {
        printf ("TCP connection broken, nr = %d, errno = %d\n", nr, errno);
        exit (EXIT_FAILURE);
      }
    }
    else if ((gdbsock = poll_socket (srvsock)) > 0) {
      print_client_info ();
      tcpok = -1;
    }

  }

  /* If we are connected on the debug port (slkport = 0)
     resend the status query. The initial status response got eaten
     in the device handshake */
  if (slkport == 0) {
    if (verbose) {
      printf("tcpusb: (Re)sending status query on debug socket\n");
    }
    send_status_query ();
  }

  /* Now our simple minded bridge */
  printf ("***** Starting Bridge *****\n");
  for (;;) {
    nr = read (gdbsock, buff, 256);
    if (nr > 0) {
      if (write_usb (buff, nr, 100) != nr) {
        /* Error on USB write */
        break;
      }
      else {
        logtraffic (logfile, '>', buff, nr);
      }
    }
    else if (nr == 0 || errno != EWOULDBLOCK) {
      /* EOF or error on TCP read, nothing to read is OK */
      printf("TCP connection broken, nr = %d, errno = %d\n", nr, errno);
      printf("Trying to send SysReset to device.\n");
      send_SysReset(0);
      send_SysReset(1);
      break;
    }

    nr = read_usb (buff, 128, 10);
    if (nr > 0 ) {
      if (write (gdbsock, buff, nr) != nr) {
        /* Error on TCP write */
        break;
      }
      else {
        logtraffic (logfile, '<', buff, nr);
      }
    }
    else if (nr < 0 && nr != -ETIMEDOUT) {
      /* Error on USB read, nothing to read is OK */
      printf("USB connection broken.\n");
      break;
    }
  }
  printf ("***** Bridge Stopped *****\n");

  /* Clean up USB side */
  close_usb ();

  /* Clean up TCP side */
  close (gdbsock);
  close (srvsock);

  return EXIT_SUCCESS;
}

