/* usbside.c: USB routines for tcpusb.

   Copyright 2005 Ton van Overbeek.

   This is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.  */

#include <stdio.h>
#include <string.h>
#include <usb.h>

/* External functions */

int rd_initial (void);
void send_FtrGet (int socket);
int rcv_FtrGet (void);
char *getromstr (void);
void logtraffic (FILE *logfile, char direction, char *buff, int size);

/* External global variables */

extern int verbose;

/* local globals for USB device */

static struct usb_device *dev = (struct usb_device *)0;
static usb_dev_handle *udev = (usb_dev_handle *)0;
static int wr_ep = -1;
static int rd_ep = -1;
static unsigned char ifacenr;

/* USB control requests we send to the devices */

#define GENERIC_REQUEST_BYTES_AVAILABLE         0x01
#define GENERIC_CLOSE_NOTIFICATION              0x02
#define VISOR_GET_CONNECTION_INFORMATION        0x03
#define PALM_GET_EXT_CONNECTION_INFORMATION     0x04

/* Portable memory access macros */

#define get_short(ptr) ((((unsigned char*)(ptr))[0] << 8)  | \
                        (((unsigned char*)(ptr))[1] << 0))

#define set_long(ptr,val) ((((unsigned char*)(ptr))[0] = ((val) >> 24) & 0xff), \
                          (((unsigned char*)(ptr))[1] = ((val) >> 16) & 0xff), \
                          (((unsigned char*)(ptr))[2] = ((val) >> 8) & 0xff), \
                          (((unsigned char*)(ptr))[3] = ((val) >> 0) & 0xff))

/* Reply  struct for PALM_GET_EXT_CONNECTION_INFORMATION */

typedef struct {
  u_int8_t num_ports;
  u_int8_t endpoint_numbers_different;
  u_int16_t reserved1;
  struct {
    u_int32_t port_function_id;
    u_int8_t port;
    u_int8_t endpoint_info;
    u_int16_t reserved;
  } connections[2];
} palm_ext_connection_info_t;

/* Reply struct and defines for VISOR_GET_CONNECTION_INFORMATION */

typedef struct {
  u_int16_t num_ports;
  struct {
    u_int8_t port_function_id;
    u_int8_t port;
  } connections[8];
} visor_connection_info_t;

/* struct visor_connection_info.connection[x].port defines: */
#define VISOR_ENDPOINT_1        0x01
#define VISOR_ENDPOINT_2        0x02

/* struct visor_connection_info.connection[x].port_function_id defines: */
#define VISOR_FUNCTION_GENERIC              0x00
#define VISOR_FUNCTION_DEBUGGER             0x01
#define VISOR_FUNCTION_HOTSYNC              0x02
#define VISOR_FUNCTION_CONSOLE              0x03
#define VISOR_FUNCTION_REMOTE_FILE_SYS      0x04

/* typedef for PalmOS device */

typedef struct palmos_usb_dev {
  u_int16_t vendor, product;
  char *idstr;          /* From known_devices table */
  char *vendorstr;      /* Read from device */
  char *productstr;     /* Read from device */
  char *serialstr;      /* Read from device */
  char *romstr;         /* Read from device with Remote Procedure Call */
} palmos_usb_dev_t;

/* Palm OS Device found */

palmos_usb_dev_t palmos_usb_dev;

/*
 * Table with known devices. The information in this table comes from the
 * corresponding table in libpisock from pilot-link (http://www.pilot-link.org)
 */

static palmos_usb_dev_t known_devices[] = {

  /* Sony */
  { .vendor = 0x054c, .product = 0x0038,
    .idstr = "Sony Palm OS 3.5 device, S300",
  },
  { .vendor = 0x054c, .product = 0x0066,
    .idstr = "Sony T, S320, SJ series, and other Palm OS 4.0 devices",
  },
  { .vendor = 0x054c, .product = 0x0095,
    .idstr = "Sony S360",
  },
  { .vendor = 0x054c, .product = 0x000a,
    .idstr = "Sony NR or other Palm OS 4.1 device",
  },
  { .vendor = 0x054c, .product = 0x00da,
    .idstr = "Sony NX",
  },
  { .vendor = 0x054c, .product = 0x00e9,
    .idstr = "Sony NZ",
  },
  { .vendor = 0x054c, .product = 0x0144,
    .idstr = "Sony UX",
  },
  { .vendor = 0x054c, .product = 0x0169,
    .idstr = "Sony TJ",
  },

  /* AlphaSmart */
  { .vendor = 0x081e, .product = 0xdf00,
    .idstr = "Alphasmart Dana",
  },

  /* HANDSPRING */
  { .vendor = 0x082d, .product = 0x0100,
    .idstr = "Handspring Visor, Treo 300",
  },
  { .vendor = 0x082d, .product = 0x0200,
    .idstr = "Handspring Treo",
  },
  { .vendor = 0x082d, .product = 0x0300,
    .idstr = "Handspring Treo 600",
  },

  /* PalmOne, Palm Inc */
  { .vendor = 0x0830, .product = 0x0001,
    .idstr = "Palm m500",
  },
  { .vendor = 0x0830, .product = 0x0002,
    .idstr = "Palm m505",
  },
  { .vendor = 0x0830, .product = 0x0003,
    .idstr = "Palm m515",
  },
  { .vendor = 0x0830, .product = 0x0010,
  },
  { .vendor = 0x0830, .product = 0x0011,
  },
  { .vendor = 0x0830, .product = 0x0020,
    .idstr = "Palm i705",
  },
  { .vendor = 0x0830, .product = 0x0030,
    .idstr = "Palm Tungsten|Z",
  },
  { .vendor = 0x0830, .product = 0x0031,
    .idstr = "Palm Tungsten|W",
  },
  { .vendor = 0x0830, .product = 0x0040,
    .idstr = "Palm m125",
  },
  { .vendor = 0x0830, .product = 0x0050,
    .idstr = "Palm m130",
  },
  { .vendor = 0x0830, .product = 0x0051,
  },
  { .vendor = 0x0830, .product = 0x0052,
  },
  { .vendor = 0x0830, .product = 0x0053,
  },
  { .vendor = 0x0830, .product = 0x0060,
    .idstr = "Palm Tungsten (T, T2, T3), Zire 71",
  },
  { .vendor = 0x0830, .product = 0x0061,
    .idstr = "Palm Zire 31, 72, Tungsten (T5, E2), Treo 650",
  },
  { .vendor = 0x0830, .product = 0x0062,
  },
  { .vendor = 0x0830, .product = 0x0063,
  },
  { .vendor = 0x0830, .product = 0x0070,
    .idstr = "Palm Zire",
  },
  { .vendor = 0x0830, .product = 0x0071,
  },
  { .vendor = 0x0830, .product = 0x0080,
    .idstr = "Palm serial adapter",
  },
  { .vendor = 0x0830, .product = 0x0099,
  },
  { .vendor = 0x0830, .product = 0x0100,
  },

  /* GARMIN */
  { .vendor = 0x091e, .product = 0x0004,
    .idstr = "Garmin IQUE 3600",
  },

  /* Kyocera */
  { .vendor = 0x0c88, .product = 0x0021,
    .idstr = "Kyocera 7135 Smartphone",
  },
  { .vendor = 0x0c88, .product = 0xa226,
    .idstr = "Kyocera 6035 Smartphone",
  },

  /* Tapwave */
  { .vendor = 0x12ef, .product = 0x0100,
    .idstr = "Tapwave Zodiac, Zodiac2",
  },

  /* ACEECA */
  { .vendor = 0x4766, .product = 0x0001,
    .idstr = "ACEECA MEZ1000",
  },

  /* Samsung */
  { .vendor = 0x04e8, .product = 0x8001,
    .idstr = "Samsung i330",
  },
};

/* Vendor defines */

#define VENDOR_TAPWAVE	0x12ef

/* Configuration reporting functions */

static void
print_endpoint (struct usb_endpoint_descriptor *endpoint)
{
  printf ("      bEndpointAddress: %02xh\n", endpoint->bEndpointAddress);
  printf ("      bmAttributes:     %02xh\n", endpoint->bmAttributes);
  printf ("      wMaxPacketSize:   %d\n", endpoint->wMaxPacketSize);
  printf ("      bInterval:        %d\n", endpoint->bInterval);
  printf ("      bRefresh:         %d\n", endpoint->bRefresh);
  printf ("      bSynchAddress:    %d\n", endpoint->bSynchAddress);
}

static void
print_altsetting (struct usb_interface_descriptor *interface)
{
  int i;

  printf ("    bInterfaceNumber:   %d\n", interface->bInterfaceNumber);
  printf ("    bAlternateSetting:  %d\n", interface->bAlternateSetting);
  printf ("    bNumEndpoints:      %d\n", interface->bNumEndpoints);
  printf ("    bInterfaceClass:    %d\n", interface->bInterfaceClass);
  printf ("    bInterfaceSubClass: %d\n", interface->bInterfaceSubClass);
  printf ("    bInterfaceProtocol: %d\n", interface->bInterfaceProtocol);
  printf ("    iInterface:         %d\n", interface->iInterface);

  for (i = 0; i < interface->bNumEndpoints; i++) {
    printf ("     --- Endpoint %d ---\n", i);
    print_endpoint (&interface->endpoint[i]);
  }
}

static void
print_interface (struct usb_interface *interface)
{
  int i;

  for (i = 0; i < interface->num_altsetting; i++) {
    printf ("    --- Altsetting %d ---\n", i);
    print_altsetting (&interface->altsetting[i]);
  }
}

static void
print_configuration (struct usb_config_descriptor *config)
{
  int i;

  printf ("------- Device Configuration -------\n");
  printf ("  wTotalLength:         %d\n", config->wTotalLength);
  printf ("  bNumInterfaces:       %d\n", config->bNumInterfaces);
  printf ("  bConfigurationValue:  %d\n", config->bConfigurationValue);
  printf ("  iConfiguration:       %d\n", config->iConfiguration);
  printf ("  bmAttributes:         %02xh\n", config->bmAttributes);
  printf ("  MaxPower:             %d\n", config->MaxPower);

  for (i = 0; i < config->bNumInterfaces; i++) {
    printf ("   --- Interface %d ---\n", i);
    print_interface (&config->interface[i]);
  }

  printf ("------------------------------------\n");
}

/* Device specific configuration */

static int
decode_generic_connection_information (palm_ext_connection_info_t *pci, int *port_no,                                               int *in_pipe_no, int *out_pipe_no)
{
  int i;

  if (verbose)
    printf ("decode_generic_connection_information: num_ports=%d, endpoint_numbers_different=%d\n",
            pci->num_ports, pci->endpoint_numbers_different);

  /* The Treo 600 returns bogus information on the debug/console connection
   * (on the 'psys' function_id, the information returned for 'sync' is OK)
   * so we hardcode the port_no. It uses the same port_no/pipes for
   * console/debug as HotSync. */
  if (palmos_usb_dev.vendor == 0x082d && palmos_usb_dev.product == 0x0300) {
    if (port_no) *port_no = 2;
    return -1;
  }

  for (i=0; i < pci->num_ports; i++) {
    char string[5];
    set_long (string, pci->connections[i].port_function_id);
    string[4] = 0;
    if (verbose) {
      printf ("\t[%d] port_function_id='%s'\n", i, string);
      printf ("\t[%d] port=%d\n", i, pci->connections[i].port);
      printf ("\t[%d] endpoint_info=0x%02x\n", i, pci->connections[i].endpoint_info);
    }
    if (pci->connections[i].port_function_id == 'psys') {
      if (!pci->endpoint_numbers_different) {
        if (port_no) *port_no = pci->connections[i].port;
      }
      else if (pci->connections[i].endpoint_info) {
        *in_pipe_no  = pci->connections[i].endpoint_info >> 4;
        *out_pipe_no = pci->connections[i].endpoint_info & 0x0f;
      }
      return -1;
    }
  }
  return 0;
}

static int
get_pipe_info (int *port_no, int *in_pipe_no, int *out_pipe_no)
{
  int i, ret;
  palm_ext_connection_info_t pci = { 0 };
  visor_connection_info_t vci = { 0 };

  /* Get palm extended connection information */
  ret = usb_control_msg (udev, 0xc2, PALM_GET_EXT_CONNECTION_INFORMATION, 0, 0,
                         (char *) &pci, sizeof (pci), 1000);
  if (verbose) {
    printf ("PALM_GET_EXT_CONNECTION_INFORMATION returns %d\n", ret);
    if ( ret > 0)
      logtraffic (stdout, 'P', (char *) &pci, ret);
  }
  if (ret <= 0) {
    /* Try visor connection info instead */
    ret = usb_control_msg (udev, 0xc2, VISOR_GET_CONNECTION_INFORMATION,  0, 0,
                           (char *) &vci, sizeof (vci), 1000);
    if (verbose) {
      printf ("VISOR_GET_CONNECTION_INFORMATION returns %d\n", ret);
      if (ret > 0)
        logtraffic (stdout, 'V', (char *) &vci, ret);
    }
    if (ret <= 0)
      return 0;

    /* Decode visor connection information */
    vci.num_ports = get_short (&vci.num_ports);	/* num_ports is little endian */
    if (vci.num_ports > 8) vci.num_ports = 8;
    if (verbose)
      printf("decode_visor_connection_information: num_ports = %d\n", vci.num_ports);
    for (i=0; i < vci.num_ports; i++) {
      if (verbose) {
        printf("\t[%d] port_function_id = 0x%02x\n", i, vci.connections[i].port_function_id);
        printf("\t[%d] port = %d\n", i, vci.connections[i].port);
      }
      if (vci.connections[i].port_function_id == VISOR_FUNCTION_DEBUGGER ||
          vci.connections[i].port_function_id == VISOR_FUNCTION_CONSOLE) {
        if (port_no) *port_no = vci.connections[i].port;
        return -1;
      }
    }
    return 0;
  }

  /* Decode palm extended connection information */
  return decode_generic_connection_information (&pci, port_no, in_pipe_no, out_pipe_no);
}

static int
find_end_points (int pipe_info, int port_no, int in_pipe_no, int out_pipe_no)
{
  /* The logic in this function follows the logic of find_interfaces() in darwinusb.c (part of
   * pilot-link http://www.pilot-link.org )
   */

  int i;
  int ret;
  struct usb_interface_descriptor *iface;
  int piperef;
  int pass;
#ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
  int first;
#endif

  /* Cycle over all the interfaces (Palm devices usually only have 1) */
  for (i = 0; i < dev->config[0].interface[0].num_altsetting; i++) {
    iface = &dev->config[0].interface[0].altsetting[i];

    /* Claim the interface */
    ifacenr = iface->bInterfaceNumber;
#ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
    first = 1;
claim:
#endif
    if ((ret = usb_claim_interface (udev, ifacenr)) < 0) {
      if (verbose)
        printf ("find_end_points: Cannot claim interface %d, usb_claim_interface() returns %d\n",
                ifacenr, ret);
#ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
      if (first) {
        usb_detach_kernel_driver_np (udev, ifacenr);
        first = 0;
        goto claim;
      }
#endif
      continue;
    }

    /* If device didn't answer to the connection information request in get_pipe_info(), try
     * to read a preamble sent by the device. This is sent by devices with chipsets that do
     * not support the vendor control endpoint requests. Look for the first pipe on which a
     * preamble is available.
     */
    if (!pipe_info) {
      int req_timeout;
      int preamblefound = 0;
      char buff[256];

      for (req_timeout = 100; req_timeout <= 300 && !preamblefound; req_timeout += 100) {
        if (verbose)
          printf("find_end_points: checking for pipe_info with timeout %dms\n", req_timeout);

        for (piperef = 0; piperef < iface->bNumEndpoints; piperef++) {
          struct usb_endpoint_descriptor *ep = &iface->endpoint[piperef];

          if ((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_DIR_MASK) {
            /* Read endpoint */

            int type = ep->bmAttributes & USB_ENDPOINT_TYPE_MASK;
            if (verbose)
              printf ("find_end_points: trying pipe %d type %d\n", piperef, type);

            if (type == USB_ENDPOINT_TYPE_BULK) {
              ret = usb_bulk_read (udev, ep->bEndpointAddress, buff, sizeof(buff), req_timeout);
              /* What about the other endpoint types with libusb ???? */
              if (ret >= 8) {
                if (verbose) {
                  printf ("find_end_points: got %d bytes!\n", ret);
                  logtraffic (stdout, '<', buff, ret);
                }
                if (!memcmp (buff, "VNDR10", 6)) {
                  palm_ext_connection_info_t pci;
                  memcpy (&pci, &buff[6], sizeof(pci));
                  decode_generic_connection_information (&pci, &port_no, &in_pipe_no, &out_pipe_no);
                  preamblefound = -1;
                  break;
                }
              }
            }            
          }
        }
      }
    }

    /* Find the endpoints we are going to use for reading and writing.
     * We have four chances to find the right endpoints:
     * 1. If we got a hint from the device, we try this first.
     * 2. If we did not get both endpoints, try using the port_no hint.
     * 3. If we are still missing one or two endpoints, give a second try looking for
     *    endpoints with a 64 byte transfer size.
     * 4. Finally if this failed, forget about the transfer size and take the first ones
     *    that come (i.e. Tungsten W has a 64 bytes IN pipe and a 32 bytes OUT pipe).
     */
    for (pass = 1; pass <= 4 && (rd_ep == -1 || wr_ep == -1); pass++) {
      
      if (verbose)
        printf("find_end_points: pass %d looking for endpoints, port_no=0x%02x,"
               " in_pipe_no=0x%02x, out_pipe_no=0x%02x\n",
               pass, port_no, in_pipe_no, out_pipe_no); 

      for (piperef = 0; piperef < iface->bNumEndpoints; piperef++) {
        struct usb_endpoint_descriptor *ep = &iface->endpoint[piperef];

        if (rd_ep == -1 &&
            (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_DIR_MASK /* read */ &&
            (ep->bmAttributes & USB_ENDPOINT_TYPE_MASK) == USB_ENDPOINT_TYPE_BULK) {
          if ((pass == 1 && in_pipe_no != 0xff && (ep->bEndpointAddress & 0x7f) == in_pipe_no) ||
              (pass == 2 && port_no != 0xff && (ep->bEndpointAddress & 0x7f) == port_no) ||
              (pass == 3 && ep->wMaxPacketSize == 64) ||
               pass == 4)
            rd_ep = ep->bEndpointAddress;
        }
        else if (wr_ep == -1 &&
            (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == 0 /* write */ &&
            (ep->bmAttributes & USB_ENDPOINT_TYPE_MASK) == USB_ENDPOINT_TYPE_BULK) {
          if ((pass == 1 && out_pipe_no != 0xff && (ep->bEndpointAddress & 0x7f) == out_pipe_no) ||
              (pass == 2 && port_no != 0xff && (ep->bEndpointAddress & 0x7f) == port_no) ||
              (pass == 3 && ep->wMaxPacketSize == 64) ||
               pass == 4)
            wr_ep = ep->bEndpointAddress;
        }
      }
    }

    if (rd_ep != -1 && wr_ep != -1) {
      if (verbose)
        printf("find_end_points: USB connection open: rd_ep=0x%02x, wr_ep=0x%02x\n", rd_ep, wr_ep);
      break;
    }

    if (verbose)
      printf("find_end_points: Could not find endpoints on this interface\n");

    usb_release_interface (udev, ifacenr);
    rd_ep = wr_ep = -1;

  }
  return ((rd_ep != -1 && wr_ep != -1) ? -1 : 0);
}

static int
configure_dev (void)
{
  static char vendorstr[256];
  static char productstr[256];
  static char serialstr[256];
  int ret;
  int pipe_info = 0;
  int port_no = 0xff;
  int in_pipe_no = 0xff;
  int out_pipe_no = 0xff;
  unsigned char ba[2] = { 0 };

  if (!dev->config) {
    if (verbose)
      printf ("configure_dev: No configuration ????\n");
    return 0;
  }
  /* We have to set the configuration here, otherwise libusb-win32 does not
   * work properly. On Linux you can get away without setting the configuration
   * explicitly. Unfortunately, setting the configuration resets the endpoints
   * in the device and anything already sent is lost.
   */
  ret = usb_set_configuration (udev, dev->config[0].bConfigurationValue);
  if (ret) {
    if (verbose)
      printf ("configure_dev: Could not set configuration %d, ret = %d\n",
              dev->config[0].bConfigurationValue, ret);
    return 0;
  }
  if (verbose)
    print_configuration (&dev->config[0]);

  if (dev->descriptor.iManufacturer && 
      (usb_get_string_simple (udev, dev->descriptor.iManufacturer, vendorstr, sizeof(vendorstr) - 1) > 0)) {
    palmos_usb_dev.vendorstr = vendorstr;
    if (verbose) {
      printf("configure_dev: vendorstr = %s\n", vendorstr);
    }
  }

  if (dev->descriptor.iProduct &&
      (usb_get_string_simple (udev, dev->descriptor.iProduct, productstr, sizeof(productstr) - 1) > 0)) {
    palmos_usb_dev.productstr = productstr;
    if (verbose) {
      printf("configure_dev: productstr = %s\n", productstr);
    }
  }

  if (dev->descriptor.iSerialNumber) {
    ret = usb_get_string_simple (udev, dev->descriptor.iSerialNumber, serialstr, sizeof(serialstr));
    if (ret > 0) {
      int i;

      if (verbose) {
        printf("configure_dev: serialstr = %s\n", serialstr);
      }

      /* Convert serialno for Palm Devices */
      if (palmos_usb_dev.vendor == 0x0830) {
        for (i=0; i < strlen(serialstr); i+=2) {
          serialstr[i/2] = (serialstr[i] - '0') * 16 + (serialstr[i+1] - '0');
        }
        serialstr[i/2] = '\0';
      }
      palmos_usb_dev.serialstr = serialstr;
    }
  }

  pipe_info = get_pipe_info (&port_no, &in_pipe_no, &out_pipe_no);

  /* Query bytes available. We do not use the information, but most devices expect to
   * receive this before talking to us
   */
  if (dev->descriptor.idVendor != VENDOR_TAPWAVE) { 
    ret = usb_control_msg (udev, 0xc2, GENERIC_REQUEST_BYTES_AVAILABLE, 0, 0, &ba[0], 2, 1000);
    if (verbose) {
      printf("configure_dev: GENERIC_REQUEST_BYTES_AVAILABLE returns %d\n", ret); 
      if (ret > 0)
        logtraffic (stdout, 'G', &ba[0], ret);
    }
    if (ret < 0)
      return 0;
  }

  return find_end_points (pipe_info, port_no, in_pipe_no, out_pipe_no);
}

/* Routines for USB bus scanning */

static int
check_device (struct usb_device *dev)
{
  unsigned int i;

  for (i=0; i < (sizeof(known_devices) / sizeof(known_devices[0])); i++) {
    if (dev->descriptor.idVendor == known_devices[i].vendor &&
        dev->descriptor.idProduct == known_devices[i].product) {
      palmos_usb_dev.vendor = known_devices[i].vendor;
      palmos_usb_dev.product = known_devices[i].product;
      palmos_usb_dev.idstr = known_devices[i].idstr;
      if (verbose)
        printf("check_device: Found device: %s\n",  palmos_usb_dev.idstr ?  palmos_usb_dev.idstr : "<NoName>");
      return -1;
    }
  }
  return 0;
}

static int
scan_busses (struct usb_device **dev)
{
  struct usb_bus *bus;
  
  usb_find_busses ();
  usb_find_devices ();

  for (bus = usb_get_busses (); bus; bus = bus->next) {
    for (*dev = bus->devices; *dev; *dev = (*dev)->next) {
      if (check_device (*dev))
        return -1;
    }
  }
  return 0;
}

/* Public funtions */

int
poll_usb (int *slkport)
{
  static int retrycount = 0;
  int ret;

start:
  if (!scan_busses (&dev))
    return 0;
  udev = usb_open (dev);

  if (!configure_dev ()) {
    fprintf (stderr, "poll_usb: Could not configure PalmOS PDA\n");
    exit (EXIT_FAILURE);
  }

  ret = rd_initial ();
  if (verbose)
    printf ("poll_usb: rd_initial() returns %d\n", ret);

  /* Try console socket first */
  send_FtrGet (1);
  ret = rcv_FtrGet ();
  if (verbose)
    printf ("poll_usb: rcv_FtrGet() con returns %d\n", ret);
  if (ret > 0) {
    palmos_usb_dev.romstr = getromstr ();
    *slkport = 1;
    return -1;
  }
  if (ret == 0) {
    /* Try again on debug socket */
    send_FtrGet (0);
    ret = rcv_FtrGet ();
    if (verbose)
      printf("poll_usb: rcv_FtrGet() dbg returns %d\n", ret);
    if (ret > 0) {
      palmos_usb_dev.romstr = getromstr ();
      *slkport = 0;
      return -1;
    }
    else {
      usb_close (udev);
      if (!retrycount++) {
        if (verbose)
          printf ("poll_usb: Retrying scan ...\n");
        goto start;
      }
    }
  }
  fprintf (stderr, "poll_usb: Could not communicate on either console or debug socket\n");
  exit (EXIT_FAILURE);
}

void
init_usb (int debuglevel)
{
  usb_set_debug (debuglevel);
  usb_init ();
}

void
close_usb (void)
{
  int ret;

  ret = usb_release_interface (udev, ifacenr);
  if (ret < 0)
    printf ("close_usb: Unable to release interface, ret = %d\n", ret);

  ret = usb_close (udev);
  if (ret < 0)
    printf ("close_usb: Unable to close device, ret = %d\n", ret);
}

int
write_usb (char *buff, int size, int timeout)
{
  return usb_bulk_write (udev, wr_ep, buff, size, timeout);
}

int
read_usb (char *buff, int size, int timeout)
{
  return usb_bulk_read (udev, rd_ep, buff, size, timeout);
}

void print_dev_info (void)
{
  if (palmos_usb_dev.idstr)
    printf ("%s\n", palmos_usb_dev.idstr);
  printf ("Vendorid = 0x%0x, Productid = 0x%0x\n", palmos_usb_dev.vendor,
          palmos_usb_dev.product);
  if (palmos_usb_dev.productstr || palmos_usb_dev.productstr ||
      palmos_usb_dev.serialstr || palmos_usb_dev.romstr) {
    printf ("Device reports:\n");
    if (palmos_usb_dev.vendorstr)
      printf ("  Vendor:  %s\n", palmos_usb_dev.vendorstr);
    if (palmos_usb_dev.productstr)
      printf ("  Product: %s\n", palmos_usb_dev.productstr);
    if (palmos_usb_dev.serialstr)
      printf ("  Serial:  %s\n", palmos_usb_dev.serialstr);
    if (palmos_usb_dev.romstr)
      printf ("  OSvers:  %s\n", palmos_usb_dev.romstr);
  }
}
